/**
FAT Filesystem Traversal routines

Orion Sky Lawlor, olawlor@acm.org, 2006/04/02 (Public Domain, except GPL where marked)
*/
#ifndef __OSL_FAT_LAYOUT_H
#define __OSL_FAT_LAYOUT_H

#include "osl/io_types.h"

typedef unsigned char byte;
typedef osl::io::Lil16 Lil16;
typedef osl::io::Lil32 Lil32;


/* Fat DISK layout:
  Boot sector and other "reserved" sectors listed in boot block
  FAT 1
  FAT 2
  Clusters of file data

*/

/** FAT boot sector layout.

Structs and #defines taken from Linux kernel: linux/include/linux/msdos_fs.h 
  --- The following is GPL code ---
*/
struct fat_boot_sector {
	byte	ignored[3];	/* Boot strap short or near jump */
	byte	system_id[8];	/* Name - can be used to special case
				   partition manager volumes */
	Lil16	sector_size;	/* bytes per logical sector */
	byte	sec_per_clus;	/* sectors/cluster */
	Lil16	reserved;	/* reserved sectors; size of boot sector stuff */
	byte	fats;		/* number of FATs */
	byte	dir_entries[2];	/* root directory entries */
	Lil16	sectors;	/* total number of sectors on disk */
	byte	media;		/* media code */
	Lil16	fat_length;	/* sectors/FAT */
	Lil16	secs_track;	/* sectors per track */
	Lil16	heads;		/* number of heads */
	Lil32	hidden;		/* hidden sectors (unused) */
	Lil32	total_sect;	/* number of sectors (if sectors == 0) */

	/* The following fields are only used by FAT32 */
	Lil32	fat32_length;	/* sectors/FAT */
	Lil16	flags;		/* bit 8: fat mirroring, low 4: active fat */
	byte	version[2];	/* major, minor filesystem version */
	Lil32	root_cluster;	/* first cluster in root directory */
	Lil16	info_sector;	/* filesystem info sector */
	Lil16	backup_boot;	/* backup boot sector */
	Lil16	reserved2[6];	/* Unused */
};

/** FAT directory entry. */
struct msdos_dir_entry {
	byte	name[8],ext[3];	/* MS-DOS name and extension */
#define DELETED_FLAG 0xe5 /* marks file as deleted when in name[0] */

	byte	attr;		/* attribute bits */
#define ATTR_NONE    0 /* no attribute bits */
#define ATTR_RO      1  /* read-only */
#define ATTR_HIDDEN  2  /* hidden */
#define ATTR_SYS     4  /* system */
#define ATTR_VOLUME  8  /* volume label */
#define ATTR_DIR     16 /* directory */
#define ATTR_ARCH    32 /* archived */

	byte    lcase;		/* Case for base and extension */
	byte	ctime_ms;	/* Creation time, milliseconds */
	Lil16	ctime;		/* Creation time */
	Lil16	cdate;		/* Creation date */
	Lil16	adate;		/* Last access date */
	Lil16   starthi;	/* High 16 bits of cluster in FAT32 */
	Lil16	time,date,start;/* time, date and first cluster */
	unsigned int start_cluster(void) const {return start+(starthi<<16);}
	Lil32	size;		/* file size (in bytes) */
};


/** Windows long filename support directory entry */
struct win_dir_entry {
	byte    id;		/* sequence number for slot */
	byte    name0_4[10];	/* first 5 characters in name */
	byte    attr;		/* attribute byte */
	byte    reserved;	/* always 0 */
	byte    alias_checksum;	/* checksum for 8.3 alias */
	byte    name5_10[12];	/* 6 more characters in name */
	Lil16   start;		/* starting cluster number, 0 in long slots */
	byte    name11_12[4];	/* last 2 characters in name */
};

/* --- END GPL CODE --- */

/** Accepts pieces of a read-in file */
class fat_consumer { 
public:
	/** Consumer this sector.  Sectors are passed in file order.
	    Return true if we want more sectors */
	virtual bool eat_sector(void *data,int len) =0;
	virtual ~fat_consumer();
};

/** Manages access to a File Allocation Table filesystem */
class fat_filesystem {
public:
	fat_filesystem();
	virtual ~fat_filesystem();
/** Read the boot sector and FAT */
	virtual void setup(void);
	
/** The filesystem's first, or "boot sector" / "superblock" */
	fat_boot_sector boot_sector;
	
	/** Disk byte offset.  A 64-bit type if possible. */
	typedef unsigned long off_t;
	/** Cluster number.  A 64-bit type if possible. */
	typedef unsigned long cluster_t;
	/** Cluster containing start of root directory entry */
	cluster_t root_cluster;
	
/** Read blocks from this file, and keep passing
   them to this consumer until it returns false (or EOF). */
   	void read_file(cluster_t first_cluster,fat_consumer &c);
	
/**
   Access to underlying disk: read n bytes into p starting at byte offset o.
   Actually implemented by a subclass of this class.
*/
	virtual void read_disk(void *p,unsigned int n,off_t o) =0;

/** Something went horribly wrong-- leave. */
	virtual void abort(const char *why);
protected:
/** The file allocation table itself */
	/** Type of one FAT entry.  FIXME: hardcoded for fat32 */
	typedef Lil32 fat_t;
	/** Links in the File Allocation Table */
	fat_t *fat; 
	/** Number of links in the FAT; also number of clusters on disk */
	off_t fat_count;
	/** Rumor has it that the high 4 bits in the FAT are reserved,
	    so mask out the high 4 bits... */
	enum {fat_mask=0x0FFFFFFF};
	/** Marks the last cluster in a file */
	enum {fat_last=0x0FFFFFF8};
	/** Return true if this is the last cluster in the FAT for this file */
	bool fat_islast(cluster_t c) {return (c&fat_mask)>=fat_last;}
	
	/** Size in bytes of each disk sector */
	off_t sector_size;
	/** Offset to first byte of first cluster */
	off_t cluster_start;
	/** Size in bytes of each disk cluster */
	off_t cluster_size;
};

#endif // def(thisHeader) 
